from __future__ import generators
from bike.parsing.extended_ast import Source
from bike.parsing.parserutils import generateLogicalLines, makeLineParseable
from bike.parsing import visitor
from bike import log
import compiler
import re

class Match:
    def __repr__(self):
        return ",".join([self.filename, self.modulefqn, str(self.lineno), str(self.colno), str(self.colend), str(self.confidence)])

def getScopeForLine(sourceNode, lineno):
    scope = None
    childnodes = sourceNode.getFlattenedListOfFastParserASTNodes()
    if childnodes == []:
        return sourceNode.fastparseroot #module node

    scope = sourceNode.fastparseroot

    for node in childnodes:
        if node.linenum > lineno: break
        scope = node

    while scope.getEndLine() <= lineno:
        scope = scope.getParent()
    return scope


def getReferences(root, matchFinder, targetname):
    for node in walkSourceNodes(root):
        print >> log.progress, "Scanning", node.filename
        if not node.containsString(targetname):
            continue
        else:
            lineno = 1
            for line in generateLogicalLines(node.fastparseroot.getMaskedLines()):
                if line.find(targetname) != -1:
                    doctoredline = makeLineParseable(line)
                    ast = compiler.parse(doctoredline)
                    scope = getScopeForLine(node, lineno)
                    matchFinder.reset(line)
                    matchFinder.setScope(scope)
                    matches = visitor.walk(ast, matchFinder).getMatches()
                    for index, confidence in matches:
                        match = Match()
                        match.filename = node.filename
                        match.modulefqn = node.fqn
                        x, y = indexToCoordinates(line, index)
                        match.lineno = lineno+y #-yoffset  
                        match.colno = x         #-xoffset
                        match.colend = match.colno+len(targetname)
                        match.confidence = confidence
                        yield match
                lineno+=line.count("\n")




def walkSourceNodes(node):
    if isinstance(node, Source):
        yield node
    else:
        for c in node.getChildNodes():
            for i in walkSourceNodes(c):
                yield i

# returns x and y coords
def indexToCoordinates(src, index):
    y = src[: index].count("\n")
    startOfLineIdx = src.rfind("\n", 0, index)+1
    x = index-startOfLineIdx
    return x, y



# interface for MatchFinder classes
# implement the visit methods
class MatchFinder:
    def setScope(self, scope):
        self.scope = scope

    def reset(self, line):
        self.matches = []
        self.words = re.split("(\w+)", line) # every other one is a non word
        self.positions = []
        i = 0
        for word in self.words:
            self.positions.append(i)
            i+=len(word)
        self.index = 0

    def getMatches(self):
        return self.matches

    def visitName(self, node):
        self.popWordsUpTo(node.name)

    def visitClass(self, node):
        self.popWordsUpTo(node.name)
        for base in node.bases:
            self.visit(base)

    def visitFunction(self, node):
        self.popWordsUpTo(node.name)
        for arg in node.argnames:
            self.popWordsUpTo(arg)
        for default in node.defaults:
            self.visit(default)
        self.visit(node.code)

    def visitGetattr(self,node):
        self.visit(node.expr)
        self.popWordsUpTo(node.attrname)

    def visitAssName(self, node):
        self.popWordsUpTo(node.name)

    def visitAssAttr(self, node):
        self.visit(node.expr)
        self.popWordsUpTo(node.attrname)

    def visitImport(self, node):
        for name, alias in node.names:
            for nameelem in name.split("."):
                self.popWordsUpTo(nameelem)
            if alias is not None:
                self.popWordsUpTo(alias)

    def visitFrom(self, node):
        for elem in ".".split(node.modname):
            self.popWordsUpTo(elem)
        for name, alias in node.names:
            self.popWordsUpTo(name)
            if alias is not None:
                self.popWordsUpTo(alias)

    def visitLambda(self, node):
        for argname in argnames:
            self.popWordsUpTo(argname)
        self.visit(defaults)
        self.visit(code)

    def visitGlobal(self, node):
        for name in names:
            self.popWordsUpTo(name)
        
    def popWordsUpTo(self, word):
        posInWords = self.words.index(word)
        idx = self.positions[posInWords]
        self.words = self.words[posInWords+1:]
        self.positions = self.positions[posInWords+1:]

    def appendMatch(self,name,confidence=100):
        idx = self.getNextIndexOfWord(name)
        self.matches.append((idx, confidence))

    def getNextIndexOfWord(self,name):
        return self.positions[self.words.index(name)]
