from __future__ import generators
from bike.query.common import Match, MatchFinder,getScopeForLine
from bike.parsing.parserutils import generateLogicalLines, makeLineParseable
import compiler
from compiler.ast import Getattr, Name
from bike.parsing.fastparserast import Class
from bike.parsing.extended_ast import Root
import re
from bike.query.getTypeOf import getTypeOfExpr, UnfoundType
from bike.parsing.visitor import walkAndGenerate

class CantFindDefinitionException:
    pass

def findDefinitionByCoords(module,lineno,col):
    node = translateSourceCoordsIntoASTNode(module,lineno,col)
    if node is None:
        raise "selected node type not supported"
    scope = getScopeForLine(module.getSourceNode(),lineno)
    type = getTypeOfExpr(scope,node)
    if type and (not isinstance(type,UnfoundType)):
        yield convertNodeToMatchObject(type,100)
    elif isinstance(node,Getattr):
        root = module.getParent()
        while not isinstance(root,Root):
            root = root.getParent()
            
        for match in scanFastParserASTForMatchingMethodNames(root,node.attrname):
            yield match

def convertNodeToMatchObject(node,confidence):
    m = Match()
    m.filename = node.module.getSourceNode().filename
    m.lineno = node.getStartLine()
    m.confidence = confidence
    return m



def translateSourceCoordsIntoASTNode(module,lineno,col):
    lline = generateLogicalLines(module.getMaskedLines()[lineno-1:]).next()
    doctoredline = makeLineParseable(lline)
    ast = compiler.parse(doctoredline)
    nodefinder = ASTNodeFinder(lline,col)
    node = compiler.walk(ast, nodefinder).node
    return node


class ASTNodeFinder(MatchFinder):
    # line is a masked line of text
    # lineno and col are coords
    def __init__(self,line,col):
        self.line = line
        self.col = col
        self.reset(line)
        self.node = None

    def visitName(self,node):
        if self.checkIfNameMatchesColumn(node.name):
            self.node = node
            return
        self.popWordsUpTo(node.name)
        
    def visitGetattr(self,node):
        self.visit(node.expr)
        if self.checkIfNameMatchesColumn(node.attrname):
            self.node = node
            return
        self.popWordsUpTo(node.attrname)

    def checkIfNameMatchesColumn(self,name):
        idx = self.getNextIndexOfWord(name)
        if idx <= self.col and idx+len(name) > self.col:
            return 1
        return 0

    def visitFrom(self, node):
        for elem in ".".split(node.modname):
            self.popWordsUpTo(elem)
        for name, alias in node.names:
            if self.checkIfNameMatchesColumn(name):
                self.node = self._manufactureASTNodeFromFQN(name)
                return
            self.popWordsUpTo(name)
            if alias is not None:
                self.popWordsUpTo(alias)


    # gets round the fact that imports etc dont contain nested getattr
    # nodes for fqns (e.g. import a.b.bah) by converting the fqn
    # string into a getattr instance    
    def _manufactureASTNodeFromFQN(self,fqn):
        if "." in fqn:
            assert 0, "getattr not supported yet"
        else:
            return Name(fqn)


def scanFastParserASTForMatchingMethodNames(root,name):
    class MethodFinder:
        def visitFunction(self,node):
            if isinstance(node.getParent(),Class):
                if node.name == name:
                    yield convertNodeToMatchObject(node,50)

    return walkAndGenerate(root,MethodFinder())


def getIndexOfWord(line,targetword):
    words = re.split("(\w+)", line)
    idx = 0
    for word in words:
        if word == targetword:
            break
        idx += len(word)
    return idx
