from __future__ import generators
from bike.parsing.extended_ast import Source
from bike.query.common import walkSourceNodes, Match, getReferences, getScopeForLine, MatchFinder
from getTypeOf import getTypeOf, getTypeOfExpr
import compiler
import re

def getReferencesToClass(root, fqn):
    classname = fqn.split(".")[-1]
    classobj = getTypeOf(root, fqn)
    classRefFinder = ClassRefFinder(classobj)
    for ref in getReferences(root, classRefFinder, classname):
        yield ref

        
class ClassRefFinder(MatchFinder):
    def __init__(self, targetclass):
        self.targetclass = targetclass

    def visitName(self, node):
        if node.name == self.targetclass.name:
            if getTypeOfExpr(self.scope, node) == self.targetclass:
                self.appendMatch(node.name)
        self.popWordsUpTo(node.name)


    def visitImport(self, node):
        for name, alias in node.names:
            if name.split(".")[-1] == self.targetclass.name:
                if getTypeOf(self.scope, name) == self.targetclass:
                    self.appendMatch(self.targetclass.name)
            for nameelem in name.split("."):
                self.popWordsUpTo(nameelem)
            if alias is not None:
                self.popWordsUpTo(alias)

    def visitGetattr(self, node):
        for c in node.getChildNodes():
            self.visit(c)
        if node.attrname == self.targetclass.name:
            if getTypeOfExpr(self.scope, node) == self.targetclass:
                self.appendMatch(self.targetclass.name)
        self.popWordsUpTo(node.attrname)

    def visitFrom(self, node):
        for elem in node.modname.split("."):
            self.popWordsUpTo(elem)
            
        for name, alias in node.names:
            if name == self.targetclass.name:
                if alias and \
                   getTypeOf(self.scope, alias) == self.targetclass:
                    self.appendMatch(self.targetclass.name)
                elif getTypeOf(self.scope, name) == self.targetclass:
                    self.appendMatch(self.targetclass.name)
            if name != "*":
                self.popWordsUpTo(name)
            if alias is not None:
                self.popWordsUpTo(alias)
