#!/usr/bin/env python
import unittest
import setpath

from bike import testdata
from bike.testutils import *
import bike
from bike.refactor.test_renameFunction import RenameFunctionTests, RenameFunctionTests_importsFunction
from bike.refactor.test_renameClass import RenameClassTests, RenameClassTests_importsClass
from bike.refactor.test_renameMethod import RenameMethodTests, RenameMethodTests_ImportsClass, RenameMethodReferenceTests, RenameMethodReferenceTests_ImportsClass, RenameMethodAfterPromptTests, TestDoesntRenameMethodIfPromptReturnsFalse
from bike.refactor import test_extractMethod
import bikefacade
from bike import UndoStackEmptyException
from bike.query.getTypeOf import getTypeOf

class TestRenameMethod(RenameMethodTests, unittest.TestCase):
    def renameMethod(self, src, fqn, newname):
        try:
            writeTmpTestFile(src)
            ctx = bike.load(tmpfile)
            ctx.renameMethod(tmpmodule+"."+fqn, newname)
            ctx.save()
            newsrc = readFile(tmpfile)
            return newsrc
        finally:
            deleteTmpTestFile()

class TestRenameMethod_importsClass(RenameMethodTests_ImportsClass, 
                                    unittest.TestCase):
    def renameMethod(self, src, fqn, newname):
        try:
            createPackageStructure(src, testdata.TheClass)
            ctx = bike.load(pkgstructureBasedir)
            ctx.renameMethod(fqn, newname)
            ctx.save()
            newsrc = readFile(os.path.join(pkgstructureBasedir, "foo.py"))
            return newsrc
        finally:
            removePackageStructure()

class TestRenameMethodReferences(RenameMethodReferenceTests, 
                                 unittest.TestCase):
    def renameMethod(self, src, fqn, newname):
        try:
            writeTmpTestFile(src)
            ctx = bike.load(tmpfile)
            ctx.renameMethod(tmpmodule+"."+fqn, newname)
            ctx.save()
            newsrc = readFile(tmpfile)
            return newsrc
        finally:
            deleteTmpTestFile()

class TestRenameMethodReferences_importsClass(RenameMethodReferenceTests_ImportsClass, 
                                              unittest.TestCase):
    def renameMethod(self, src, fqn, newname):
        try:
            createPackageStructure(src, testdata.TheClass)
            ctx = bike.load(pkgstructureBasedir)
            ctx.renameMethod(fqn, newname)
            ctx.save()
            newsrc = readFile(os.path.join(pkgstructureBasedir, "foo.py"))
            return newsrc
        finally:
            removePackageStructure()


class TestRenameMethodAfterPrompt(RenameMethodAfterPromptTests, 
                                  unittest.TestCase):
    def callback(self, filename, line, colstart, colend):
        return 1

    def renameMethod(self, src, fqn, newname):
        try:
            writeTmpTestFile(src)
            ctx = bike.load(tmpfile)
            ctx.setRenameMethodPromptCallback(self.callback)
            ctx.renameMethod(tmpmodule+"."+fqn, newname)
            ctx.save()
            newsrc = readFile(tmpfile)
            return newsrc
        finally:
            deleteTmpTestFile()

class TestDoesntRenameMethodIfPromptReturnsFalse(unittest.TestCase,
                                                 TestDoesntRenameMethodIfPromptReturnsFalse):

    def callback(self, filename, line, colstart, colend):
        return 0

    def renameMethod(self, src, fqn, newname):
        try:
            writeTmpTestFile(src)
            ctx = bike.load(tmpfile)
            ctx.setRenameMethodPromptCallback(self.callback)
            ctx.renameMethod(tmpmodule+"."+fqn, newname)
            ctx.save()
            newsrc = readFile(tmpfile)
            return newsrc
        finally:
            deleteTmpTestFile()



class TestRenameFunction(RenameFunctionTests, unittest.TestCase):
    def renameFunction(self, src, fqn, newname):
        try:
            writeTmpTestFile(src)
            ctx = bike.load(tmpfile)
            ctx.renameFunction(tmpmodule+"."+fqn, newname)
            ctx.save()
            newsrc = readFile(tmpfile)
            return newsrc
        finally:
            deleteTmpTestFile()

class TestRenameFunction_importsFunction(RenameFunctionTests_importsFunction, 
                                   unittest.TestCase):
    def renameFunction(self, src, fqn, newname):
        try:
            createPackageStructure(src, testdata.Function)
            ctx = bike.load(pkgstructureBasedir)
            ctx.renameFunction(fqn, newname)
            ctx.save()
            newsrc = readFile(os.path.join(pkgstructureBasedir, "foo.py"))
            return newsrc
        finally:
            removePackageStructure()

class TestRenameClass(RenameClassTests, unittest.TestCase):
    def renameClass(self, src, fqn, newname):
        try:
            writeTmpTestFile(src)
            ctx = bike.load(tmpfile)
            ctx.renameClass(tmpmodule+"."+fqn, newname)
            ctx.save()
            newsrc = readFile(tmpfile)
            return newsrc
        finally:
            deleteTmpTestFile()


class TestRenameClass_importsClass(RenameClassTests_importsClass, 
                                   unittest.TestCase):
    def renameClass(self, src, fqn, newname):
        try:
            createPackageStructure(src, testdata.TheClass)
            ctx = bike.load(pkgstructureBasedir)
            ctx.renameClass(fqn, newname)
            ctx.save()
            newsrc = readFile(os.path.join(pkgstructureBasedir, "foo.py"))
            return newsrc
        finally:
            removePackageStructure()

class TestRenameByCoordinates(RenameFunctionTests,
                              RenameClassTests,
                              RenameMethodTests,
                              unittest.TestCase):

    def getLineOfNode(self,ctx,fqn):
        node = getTypeOf(ctx._getAST(),tmpmodule+"."+fqn)
        return node.getStartLine()
        #node = ctx._getAST().typeinfo.getTypeOf(tmpmodule+"."+fqn)
        #return node.getStartCoords().x,node.getStartCoords().y
    
    def renameMethod(self, src, fqn, newname):
        try:
            writeTmpTestFile(src)
            ctx = bike.load(tmpfile)
            line = self.getLineOfNode(ctx,fqn)
            ctx.renameByCoordinates(os.path.abspath(tmpfile),line,newname)
            ctx.save()
            newsrc = readFile(tmpfile)
            return newsrc
        finally:
            deleteTmpTestFile()

    renameFunction = renameMethod
    renameClass = renameMethod


    def test_raisesExceptionIfCantFindNodeFromCoordinates(self):
        try:
            writeTmpTestFile("pass")
            ctx = bike.load(tmpfile)
            try:
                ctx.renameByCoordinates(os.path.abspath(tmpfile),1,"NewName")
                self.fail("shouldnt get to here")
            except bike.CouldntLocateASTNodeFromCoordinatesException:
                pass
        finally:
            deleteTmpTestFile()
        



class TestLocateNode(unittest.TestCase):
    def test_findsFunctioninNestedFile(self):
        src = testdata.Function[1:]
        createPackageStructure("pass", src)
        ctx = bike.load(pkgstructureBasedir)
        removePackageStructure()
        filename = os.path.join(os.path.abspath(pkgstructureBasedir), "b", "bah.py")
        assert ctx._locateNode(filename, 1).fqn == "a.b.bah.theFunction"

    def test_returnsNoneIfCantFindNode(self):
        src = testdata.Function[1:]
        createPackageStructure("pass", src)
        ctx = bike.load(pkgstructureBasedir)
        removePackageStructure()
        filename = os.path.join(os.path.abspath(pkgstructureBasedir), "b", "bah.py")
        assert ctx._locateNode(filename, 2) == None

    def test_findsClassinNestedFile(self):
        src = trimLines("""
        class TheClass:
            pass
            
        raise TheClass, \"hello mum\"
        """)
        createPackageStructure("pass", src)
        ctx = bike.load(pkgstructureBasedir)
        removePackageStructure()
        filename = os.path.join(os.path.abspath(pkgstructureBasedir), "b", "bah.py")
        assert ctx._locateNode(filename, 1).fqn == "a.b.bah.TheClass"


class TestReloadSource(unittest.TestCase):
    def test_changesASTToReflectNewSource(self):
        createPackageStructure("pass", "pass")
        ctx = bike.load(pkgstructureBasedir)
        removePackageStructure()
        createPackageStructure("pass", "class TheClass: pass")
        filename = os.path.join(os.path.abspath(pkgstructureBasedir), "b", "bah.py")
        ctx.load(filename)        
        removePackageStructure()
        assert ctx._locateNode(filename,1).fqn == "a.b.bah.TheClass"

    def test_canRenameClassOnNewReloadedCode(self):
        srcBefore=trimLines("""
        import b.bah.TheClass

        def foo():
            a = b.bah.TheClass()
            a.theMethod()
        """)
        srcAfter=trimLines("""
        import b.bah.NewName

        def foo():
            a = b.bah.NewName()
            a.theMethod()
        """)

        createPackageStructure(srcBefore, "pass")
        ctx = bike.load(pkgstructureBasedir)
        removePackageStructure()
        createPackageStructure("pass", "class TheClass: pass")
        filename = os.path.join(os.path.abspath(pkgstructureBasedir), "b", "bah.py")
        ctx.load(filename)
        ctx.renameClass("a.b.bah.TheClass","NewName")
        removePackageStructure()
        src = getMatchingASTNode(ctx._getAST(), "Source",fqn="a.foo").getSource()
        self.assertEqual(srcAfter,src)
        #assert getASTNode(ctx.ast,"CallFunc").node.attrname == "NewName"


class Test_deducePackageOfFile(unittest.TestCase):
    def test_returnsEmptyStringIfFileNotInPackage(self):
        try:
            # this doesnt have __init__.py file, so
            # isnt package
            os.makedirs("a")
            writeFile(os.path.join("a","foo.py"),"pass")
            pkg = bikefacade._deducePackageOfFile(os.path.join("a","foo.py"))
            assert pkg == ""
        finally:
            os.remove(os.path.join("a","foo.py"))
            os.removedirs(os.path.join("a"))

    def test_returnsNestedPackage(self):
        try:
            os.makedirs(os.path.join("a","b"))
            writeFile(os.path.join("a","__init__.py"),"# ")
            writeFile(os.path.join("a","b","__init__.py"),"# ")
            writeFile(os.path.join("a","b","foo.py"),"pass")
            pkg = bikefacade._deducePackageOfFile(os.path.join("a","b","foo.py"))
            assert pkg == "a.b"
        finally:
            os.remove(os.path.join("a","__init__.py"))
            os.remove(os.path.join("a","b","__init__.py"))
            os.remove(os.path.join("a","b","foo.py"))
            os.removedirs(os.path.join("a","b"))
        

class TestRenameMethodClientCallback(unittest.TestCase):

    class MockCallback(Mock):
        def callback(self, filename, line, colstart, colend):
            return 1
    
    def test_callbackGetsPassedTheFilenameCorrectly(self):
        src = trimLines("""
        class TheClass:
            def theMethod(self):
                pass

        a = myunfindableclass()

        a.theMethod()
        """)
        writeTmpTestFile(src)
        ctx = bike.load(tmpfile)
        deleteTmpTestFile()
        m = TestRenameMethodClientCallback.MockCallback()
        ctx.setRenameMethodPromptCallback(m.callback)
        ctx.renameMethod(tmpmodule+".TheClass.theMethod","NewName")
        assert m.mockCalledMethods["callback"][0].getParam(0) == os.path.abspath(tmpfile)

    def test_callbackGetsPassedTheCoordsCorrectly(self):
        writeTmpTestFile(trimLines("""
        class TheClass:
            def theMethod(self):
                pass
            
        a = myunfindableclass()
        
        a.theMethod()
        """))

        ctx = bike.load(tmpfile)
        deleteTmpTestFile()
        m = TestRenameMethodClientCallback.MockCallback()
        ctx.setRenameMethodPromptCallback(m.callback)
        ctx.renameMethod(tmpmodule+".TheClass.theMethod","NewName")
        self.assertEqual(m.mockCalledMethods["callback"][0].getParam(1),7)
        self.assertEqual(m.mockCalledMethods["callback"][0].getParam(2),2)
        self.assertEqual(m.mockCalledMethods["callback"][0].getParam(3),11)

class TestLoad(unittest.TestCase):
    def test_throwsExceptionIfTryToLoadANonPythonFile(self):
        try:
            writeFile("aoeuaoeuaoeu","hello")
            try:
                bike.load("aoeuaoeuaoeu")
                self.fail("should get to here")
            except bike.NotAPythonModuleOrPackageException:
                pass
        finally:
            os.remove("aoeuaoeuaoeu")

class TestExtractMethod(test_extractMethod.TestExtractMethod):

    def test_extractsPass(self):
        srcBefore=trimLines("""
        class MyClass:
            def myMethod(self):
                pass
        """)

        srcAfter=trimLines("""
        class MyClass:
            def myMethod(self):
                self.newMethod()

            def newMethod(self):
                pass
        """)

        try:
            writeTmpTestFile(srcBefore)
            ctx = bike.load(tmpfile)
            ctx.extractMethod(os.path.abspath(tmpfile),3,8,3,12,"newMethod")
            ctx.save()
            self.assertEqual(readTmpTestFile(),srcAfter)
            ctx.undo()
            ctx.save()
            self.assertEqual(readTmpTestFile(),srcBefore)
        finally:
            deleteTmpTestFile()
        

class TestExtractFunction(test_extractMethod.TestExtractFunction):
    def test_extractsFunction(self):        
        srcBefore=trimLines("""
        def myFunction(): # comment
            a = 3
            c = a + 99
            b = c * 1
            print b
        """)
        srcAfter=trimLines("""
        def myFunction(): # comment
            a = 3
            b = newFunction(a)
            print b

        def newFunction(a):
            c = a + 99
            b = c * 1
            return b
        """)
        try:
            writeTmpTestFile(srcBefore)
            ctx = bike.load(tmpfile)
            ctx.extractMethod(os.path.abspath(tmpfile),3,4,4,13,"newFunction")
            ctx.save()
            self.assertEqual(readTmpTestFile(),srcAfter)
            ctx.undo()
            ctx.save()
            self.assertEqual(readTmpTestFile(),srcBefore)
        finally:
            deleteTmpTestFile()


class TestUndo(unittest.TestCase):
    def test_undoesTheTextOfASingleFile(self):
        try:
            src = trimLines("""
            class a:
                def foo(self):
                    pass
            """)
            writeTmpTestFile(src)
            ctx = bike.load(tmpfile)
            ctx.renameMethod(tmpmodule+".a.foo","c")
            ctx.save()
            ctx.undo()
            ctx.save()
            newsrc = readFile(tmpfile)
            self.assertEqual(newsrc,src)
        finally:
            deleteTmpTestFile()
        

    def test_undoesTwoConsecutiveRefactorings(self):
        try:
            src = trimLines("""
            class a:
                def foo(self):
                    pass
            """)
            writeTmpTestFile(src)
            ctx = bike.load(tmpfile)
            ctx.renameMethod(tmpmodule+".a.foo","c")
            ctx.save()
            
            newsrc1 = readFile(tmpfile)

            ctx.renameMethod(tmpmodule+".a.c","d")
            ctx.save()

            
            # 1st undo
            ctx.undo()
            ctx.save()
            newsrc = readFile(tmpfile)
            self.assertEqual(newsrc,
                             newsrc1)

            # 2nd undo
            ctx.undo()
            ctx.save()
            newsrc = readFile(tmpfile)
            self.assertEqual(newsrc,src)
        finally:
            deleteTmpTestFile()
        

    def test_undoesTheTextOfAFileTwice(self):
        for i in range(3):
            try:
                src = trimLines("""
                class foo:
                    def bah(self):
                        pass
                """)
                writeTmpTestFile(src)
                ctx = bike.load(tmpfile)
                ctx.renameMethod(tmpmodule+".foo.bah","c")
                ctx.save()
                ctx.undo()
                ctx.save()
                newsrc = readFile(tmpfile)
                self.assertEqual(newsrc,src)
                raisedexception=0
                try:
                    ctx.undo()
                except UndoStackEmptyException:
                    raisedexception=1
                assert raisedexception
            finally:
                deleteTmpTestFile()

    def test_undoesManualModificationsToFiles(self):
        try:
            writeTmpTestFile("class foo: pass")
            origsrc = readFile(tmpfile)
            ctx = bike.load(tmpfile)

            deleteTmpTestFile()
            writeTmpTestFile("pass")
            import os
            ctx.load(tmpfile)
            newsrc = readFile(tmpfile)
            assert newsrc != origsrc
            ctx.undo()
            ctx.save()
            newsrc = readFile(tmpfile)
            assert newsrc == origsrc
        finally:
            deleteTmpTestFile()


class TestGetFullyQualifiedNameOfScope(unittest.TestCase):
    def test_getsName(self):
        src = trimLines("""
        class TheClass:
            pass
        a = TheClass()
        """)
        try:
            createPackageStructure("pass",src)
            ctx = bike.load(pkgstructureBasedir)
            self.assertEqual(ctx.getFullyQualifiedNameOfScope(pkgstructureFile2,1),"a.b.bah.TheClass")
        finally:
            removePackageStructure()

class TestgetTypeOfExpression(unittest.TestCase):
    def test_getsTypeOfSimpleClassInstanceReference(self):
        src = trimLines("""
        from b.bah import TheClass
        a = TheClass()
        a.theMethod()
        """)
        try:
            createPackageStructure(src,testdata.TheClass)
            ctx = bike.load(pkgstructureBasedir)
            self.assertEqual(ctx.getTypeOfExpression(pkgstructureFile1,3,0,11).fqn,
                             "a.b.bah.TheClass.theMethod")
        finally:
            removePackageStructure()

    def test_getsTypeIfCoordsTheWrongWayRound(self):
        src = trimLines("""
        from b.bah import TheClass
        a = TheClass()
        a.theMethod()
        """)
        try:
            createPackageStructure(src,testdata.TheClass)
            ctx = bike.load(pkgstructureBasedir)
            self.assertEqual(ctx.getTypeOfExpression(pkgstructureFile1,3,11,0).fqn,
                             "a.b.bah.TheClass.theMethod")
        finally:
            removePackageStructure()



class TestGetReferencesToClass(unittest.TestCase):
    def test_returnsReferences(self):        
        src = trimLines("""
        class TheClass:
            pass
        a = TheClass()
        """)
        try:
            writeTmpTestFile(src)
            ctx = bike.load(tmpfile)
            refs = ctx.getReferencesToClass(tmpmodule + ".TheClass")
            self.assertEqual(refs[0].filename,os.path.abspath(tmpfile))
            self.assertEqual(refs[0].lineno,3)
            assert hasattr(refs[0],"confidence")
        finally:
            deleteTmpTestFile()


class TestFindDefinitionByCoordinates(unittest.TestCase):
    def test_findsClassRef(self):
        src=trimLines("""
        class TheClass:
            pass
        a = TheClass()
        """)
        try:
            writeTmpTestFile(src)
            ctx = bike.load(tmpfile)
            defn = [x for x in
                    ctx.findDefinitionByCoordinates(tmpfile,3,6)]
            assert defn[0].filename == os.path.abspath(tmpfile)
            assert defn[0].lineno == 1
            assert defn[0].confidence == 100
        finally:
            deleteTmpTestFile()

if __name__ == "__main__":
    unittest.main()
