from bike.parsing.undo import getUndoStack
import re

# This class maintains a set of changed lines to the original source
# nodes. This is important because the act of changing a line messes
# up the coordinates on which renames are done.
# Commit writes the changes back to the source nodes
class WordRewriter:
    def __init__(self):
        self.modifiedsrc = {}

    def rewriteString(self, srcnode, lineno, colno, newname):
        if not self.modifiedsrc.has_key(srcnode):
            getUndoStack().addSource(srcnode,srcnode.getSource())
            self.modifiedsrc[srcnode] = {}
        if not self.modifiedsrc[srcnode].has_key(lineno):
            line = srcnode.getLines()[lineno-1]
            self.modifiedsrc[srcnode][lineno] = self._lineToDict(line)
        self.modifiedsrc[srcnode][lineno][colno] = newname


    # writes all the changes back to the src nodes
    def commit(self):
        for srcnode in self.modifiedsrc.keys():
            #print "srcnode.filename",srcnode.filename
            for lineno in self.modifiedsrc[srcnode]:
                srcnode.getLines()[lineno-1] = self._dictToLine(self.modifiedsrc[srcnode][lineno])
            #print "committing:\n","".join(srcnode.getLines())
            srcnode.reload("".join(srcnode.getLines()))
            srcnode.setModified()

        

    # this function creates a dictionary with each word referenced by
    # its column position in the original line
    def _lineToDict(self, line):
        words = re.split("(\w+)", line)
        h = {};i = 0
        for word in words:
            h[i] = word
            i+=len(word)
        return h

    def _dictToLine(self, d):
        cols = d.keys()
        cols.sort()
        return "".join([d[colno]for colno in cols])
