;
;   XEmacs support for bicycle repair man
;   -------------------------------------
;   Author: Phil Dawes (pdawes@users.sourceforge.net)
;
;   Eval this buffer, then use python-mode. A refactoring menu will
;   appear
;
;   If things go wrong, check the 'Bicycle Repair Man' buffer for
;   trace

(require 'python-mode)

(defvar brm-py-command nil "override this if emacs can't find the python executable")
(defvar brm-py-script nil "override this if emacs can't find the bicycle repair man script")

(defvar brm-menu nil "Menu for Python Mode.  This menu will get
created automatically if you have the `easymenu' package.  Note that
the latest X/Emacs releases contain this package.")

(defconst brm-is-gnu-emacs (string-match "GNU" (emacs-version)) "variable which says whether this is gnu-emacs or xamcs")

(and (py-safe (require 'easymenu) t)
	 (easy-menu-define
      brm-menu py-mode-map "Bicycle Repair Man "
      '("BicycleRepairMan"
				   ["Import Code Into BRM" brm-load t]
				   "---"
				   "Queries"
				   ["Get-Scope-Name" brm-get-scope-name]
				   ["Find-References" brm-find-references]
				   ["Find-Definition" brm-find-definition]
		           "---"
				   "Refactoring"
				   ["Rename-Method" brm-rename-method t]
				   ["Rename-Function" brm-rename-function t]
				   ["Rename-Class" brm-rename-class t]
				   ["Extract-Method" brm-extract-method t]
				   ["Extract-Function" brm-extract-function t]
				   ["Undo Last Refactoring" brm-undo t]

	)))

(if brm-menu
	(add-hook 'python-mode-hook (lambda () (easy-menu-add brm-menu))))


(defun brm-start ()
"starts the bicycle repair man process"
  (if brm-py-script
	  (make-comint "Bicycle Repair Man" brm-py-command nil brm-py-script)
	(make-comint "Bicycle Repair Man" "bikeemacs"))
  (set-process-filter (get-process "Bicycle Repair Man") `brm-process-filter) 
)

(defun brm-kill ()
"kills the current bicycle repair man process"
  (interactive)
  (if (get-process "Bicycle Repair Man")
	  (progn
		(kill-buffer (process-buffer (get-process "Bicycle Repair Man")))
		(kill-buffer (get-buffer-create "BicycleRepairManConsole"))))
)


(defun brm-save-buffers ()
  (if (get-process "Bicycle Repair Man")
	  (process-send-string "Bicycle Repair Man" "disableMessages\n"))
  (save-some-buffers)
  (if (get-process "Bicycle Repair Man")
	  (process-send-string "Bicycle Repair Man" "enableMessages\n"))
)


(defun brm-load (path)
"Loads the python file or directory (package) structure into bicycle repair man"
  (interactive "fPython file or directory: ")
  (brm-save-buffers)
  (if (not (get-process "Bicycle Repair Man")) (brm-start))
  (process-send-string "Bicycle Repair Man" (format "load \"%s\"\n" path))
)

(defun brm-rename-function (newname)
"Renames the function at point"
  (interactive "sRename Function To: ")
  (brm-rename newname))

(defun brm-rename-class (newname)
"Renames the class at point"
  (interactive "sRename Class To: ")
  (brm-rename newname))

(defun brm-rename-method (newname)
"Renames the method at point"
  (interactive "sRename Method To: ")
  (brm-rename newname))

(defun brm-rename (newname)
"Renames the class, function or method at point"
  (brm-save-buffers)
  (process-send-string "Bicycle Repair Man" 
                       (format "rename \"%s\" %d %s\n" 
                               (buffer-file-name)
                               (count-lines 1 (point))
                               newname)))

(defun brm-extract-method (methodname)
"Extracts the selected region into a new method"
  (interactive "sMethod name: ")
  (brm-save-buffers)
  (exchange-point-and-mark)  
  (save-excursion
	(setq beginx (current-column))
	(setq beginy (count-lines 1 (point)))
	(if (= (current-column) 0) (setq beginy (+ beginy 1)))  ; get round 'if col is 0 then line is 1 less'
	(exchange-point-and-mark)  
	(setq endx (current-column))
	(setq endy (count-lines 1 (point)))
	(if (= (current-column) 0) (setq endy (+ endy 1)))  ; get round 'if col is 0 then line is 1 less'
	(process-send-string "Bicycle Repair Man" 
                             (format "extractMethod \"%s\" %d %d %d %d %s\n" 
                                     (buffer-file-name)
                                     beginy beginx endy endx methodname))))

(defun brm-extract-function (functionname)
"Extracts the selected region into a new function"
  (interactive "sFunction name: ")
  (brm-save-buffers)
  (exchange-point-and-mark)  
  (save-excursion
	(setq beginx (current-column))
	(setq beginy (count-lines 1 (point)))
	(if (= (current-column) 0) (setq beginy (+ beginy 1)))  ; get round 'if col is 0 then line is 1 less'
	(exchange-point-and-mark)  
	(setq endx (current-column))
	(setq endy (count-lines 1 (point)))
	(if (= (current-column) 0) (setq endy (+ endy 1)))  ; get round 'if col is 0 then line is 1 less'

	(process-send-string "Bicycle Repair Man" 
                             (format "extractFunction \"%s\" %d %d %d %d %s\n" 
                                     (buffer-file-name)
                                     beginy beginx endy endx functionname))))

(defun brm-undo ()
"Undoes the last refactoring. WARNING: this is dangerous if \
the user has modified files since the last refactoring"
  (interactive)
  (process-send-string "Bicycle Repair Man" 
					   (format "undo\n")))

(defun brm-get-scope-name ()
"Prints the fully qualified name of the scope"
  (interactive)
  (process-send-string "Bicycle Repair Man" 
                       (format "getFullyQualifiedNameOfScope \"%s\" %d %d\n"
                               (buffer-file-name)
                               (count-lines 1 (point))
                               (current-column)))
)


(defun brm-find-references ()
"displays a buffer of references to the class, function or method"
  (interactive)
  (brm-save-buffers)
  (setq filename (buffer-file-name))
  (setq linenum (count-lines 1 (point)))
  (process-send-string "Bicycle Repair Man" 
                       (format "findReferences \"%s\" %d\n"
                               filename
                               linenum))
)

(defun brm-clear-console ()
  "clears the console buffer (creating it if required)"
  (switch-to-buffer-other-window (get-buffer-create "BicycleRepairManConsole"))
  (compilation-mode "BicycleRepairMan")
  (erase-buffer)
  (insert "Bicycle Repair Man\n")
  (insert "(Hint: Press Return a Link)\n")
)


(defun brm-find-definition ()
"finds the definition of the reference at point"
  (interactive)
  (brm-save-buffers)
  (setq filename (buffer-file-name))
  ;(setq begincol (current-column))
  (setq begincol (brm-get-column))
  (setq line (count-lines 1 (point)))
  (if (= begincol 0) (setq line (+ line 1)))  ; get round 'if col is 0 then line is 1 less'
  (process-send-string "Bicycle Repair Man" 
                       (format "findDefinition \"%s\" %d %d\n"
                               filename
                               line
                               begincol))
)

(defvar brm-result nil)

(defun brm-process-filter (proc string)
"Internal filter for the bicycle repair man process"
  (save-excursion
	(set-buffer (process-buffer (get-process "Bicycle Repair Man")))
	(insert string)
	)
  
  (if (string-match "BRMError: \\(.*\\)$" string)
	  (progn
		(error (match-string 1 string))))
  (if (string-match "BRMMessage: \\(.*\\)$" string)
	  (progn
		(message (match-string 1 string))
		(process-send-string "Bicycle Repair Man" "OK\n")))
  (if (string-match "BRMClearConsole" string)
	  (progn
		(brm-clear-console)
		(process-send-string "Bicycle Repair Man" "OK\n")))
  (if (string-match "BRMConsole: \\(.*\\)$" string)
	  (save-excursion
		(switch-to-buffer (get-buffer-create "BicycleRepairManConsole"))
		(end-of-buffer)
		(insert (concat (match-string 1 string) "\n"))
		(process-send-string "Bicycle Repair Man" "OK\n")))
  (if (string-match "BRMUpdated: \\(.*\\)$" string)
	  (progn
		(brm-revert-file-buffer (match-string 1 string))
		(process-send-string "Bicycle Repair Man" "OK\n")))
  (if (string-match "BRMResult: \\(.*\\)$" string)
	  (progn
		(setq brm-result (match-string 1 string))
		(process-send-string "Bicycle Repair Man" "OK\n")))
  (if (string-match "BRMGotoLine: \\([^ ]+\\) \\([^ ]+\\)$" string)
	  (progn
		(brm-goto-line (match-string 1 string)
			  (string-to-number (match-string 2 string)))
		(process-send-string "Bicycle Repair Man" "OK\n")))
  (if (string-match "BRMPromptForRename: \\([^ ]+\\) \\([^ ]+\\) \\([^ ]+\\) \\([^ ]+\\) \\([^ ]+\\)$" string)
	  (if (brm-prompt-for-rename (match-string 1 string)
								 (string-to-number (match-string 2 string))
								 (string-to-number (match-string 3 string))
								 (string-to-number (match-string 4 string))
								 (string-to-number (match-string 5 string))
								 "Couldn't deduce object type - rename this method reference? ")
		  (process-send-string "Bicycle Repair Man" "y\n")
		(process-send-string "Bicycle Repair Man" "n\n")))
)

(defun brm-prompt-for-rename (filename startx starty endx endy message)
"Internal function - goes to file and prompts user, asking if need to rename"
  (setq mybuf (current-buffer))
  (setq ans (save-excursion
			  (setq brm-updating-buffer filename)
			  (find-file filename)
			  (setq brm-updating-buffer "")
			  (goto-line starty)
			  (move-to-column startx)
			  (set-mark (+ (point) (- endx startx)))
			  (if (not brm-is-gnu-emacs) (activate-region))
			  (y-or-n-p message)))
  (switch-to-buffer mybuf)
  ans
  )


(defun brm-get-column ()
"returns the selected column. If a region has been selected, returns the start of the region"
  (interactive)
  (setq ans (save-excursion
			  (if (and (mark) (< (mark) (point)))
				  (exchange-point-and-mark))
			  (current-column))
  )
  ans
)

(defvar brm-updating-buffer "" "Internal variable containing name of file to be updated, so that it doesnt get reloaded into brm")


(defun brm-revert-file-buffer (filename)
"reverts the buffer currently looking at filename, if it exists"
  (setq mybuf (find-buffer-visiting filename))
  (if mybuf
	  (progn (set-buffer mybuf)
			 (setq brm-updating-buffer filename)
			 (revert-buffer nil 1)
			 (setq brm-updating-buffer "")
			 )
	(message "Couldnt find buffer")
	)
  )

(defun brm-goto-line (filename lineno)
  "Sets point at lineno visiting filename"
  (setq brm-updating-buffer filename)
  (find-file-other-window filename)
  (setq brm-updating-buffer "")
  (goto-line lineno)
  )

(defun brm-reload-file ()
  (if (and (string-match ".py$" (buffer-file-name)) 
           (not (brm-filename-is-currently-being-updated buffer-file-name)))
      (progn
        (if (not (get-process "Bicycle Repair Man")) (brm-start))
        (process-send-string "Bicycle Repair Man"		   
                             (format "load \"%s\"\n" (buffer-file-name)))
        )
    )
  )

(defun brm-filename-is-currently-being-updated (filename)
  (or (equal brm-updating-buffer filename) 
	  (and brm-is-gnu-emacs (equal (replace-regexp-in-string "\\\\" "/" brm-updating-buffer) filename)))
  )

(add-hook 'after-save-hook 'brm-reload-file)
(add-hook 'find-file-hooks 'brm-reload-file)


